require 'rack/handler'
require 'rack/server'
require 'rack/builder'
require 'uu/os/util/embedded_http_server'

module UU
  module OS
    module Util

      # Simple application server based on {UU::OS::Util::EmbeddedHTTPServer EmbeddedHTTPServer}
      # to be used mainly for testing purposes.
      # @example Basic usage:
      #  require 'uu/os/util/embedded_application_server'
      #
      #  server = UU::OS::Util::EmbeddedApplicationServer.new('path/to/application')
      #  server.start
      #  ...
      #  server.shutdown
      class EmbeddedApplicationServer < UU::OS::Util::EmbeddedHTTPServer

        # Creates new instance of server.
        # @param [String] app_dir Root directory of application (directory which contains
        #    configuration file +config.ru+, defaults to current working directory).
        # @param [Hash] options See {UU::OS::Util::EmbeddedHttpServer#initialize}.
        # @option options [String, #read, Hash] :server_cfg Additional configuration to
        #    be passed to application server (this configuration is merged with one given
        #    via SERVER_CFG environment variable). Possible values are:
        #
        #    * +String+ - Path to configuration file in properties format.
        #    * +#read+ - Instance of existing +IO+ (e.g. +File+) pointing to data in properties format.
        #    * +Hash+ - Configuration in plain key-value format.
        def initialize(app_dir = Dir.pwd, options = {})
          app_dir, options = Dir.pwd, app_dir if app_dir.kind_of?(Hash)
          super(options)
          @app_dir = File.expand_path(app_dir)
          @options = options.dup
          @options[:config] ||= './config.ru'
          @options[:server] ||= 'embedded'
          @options[:server_instance] = @server
          Rack::Handler.register('embedded', EmbeddedHandler)
        end

        # Starts server.
        # @return [void]
        def start
          @thread = Thread.new {
            Thread.current.abort_on_exception = true
            Thread.current[:BASE_URL] = base_url
            Thread.current[:SERVER_CFG] = @options[:server_cfg]
            Dir.chdir(@app_dir)
            Rack::Server.start(@options)
          }
          while @server.status != :Running
            Thread.pass
            unless @thread.alive?
              @thread.join
              raise
            end
          end
          return
        end

        private

        # Implementation of {Rack:Handler} for running application
        # on {UU::OS::Util::EmbeddedHttpServer} with support for
        # request streaming without buffering.
        # @private
        class EmbeddedHandler < Rack::Handler::WEBrick # :nodoc:
          def self.run(app, options = {})
            @server = options[:server_instance]
            @server.mount '/', self, app
            yield @server if block_given?
            @server.start
          end

          # Override original method to support streaming of request and response
          # (need to be done as copy-paste because it is not possible to only wrap
          # original method which call StringIO.new(req.body.to_s)) and has response
          # body as String.
          def service(req, res)
            res.rack = true
            env = req.meta_vars
            env.delete_if { |k, v| v.nil? }

            env.update({'rack.version' => Rack::VERSION,
              'rack.input' => req.body_stream,
              'rack.errors' => $stderr,

              'rack.multithread' => true,
              'rack.multiprocess' => false,
              'rack.run_once' => false,

              'rack.url_scheme' => ['yes', 'on', '1'].include?(env['HTTPS']) ? 'https' : 'http',

              'rack.hijack?' => true,
              'rack.hijack' => lambda { raise NotImplementedError, 'only partial hijack is supported.'},
              'rack.hijack_io' => nil,
            })

            env['HTTP_VERSION'] ||= env['SERVER_PROTOCOL']
            env['QUERY_STRING'] ||= ''
            unless env['PATH_INFO'] == ''
              path, n = req.request_uri.path, env['SCRIPT_NAME'].length
              env['PATH_INFO'] = path[n, path.length-n]
            end
            env['REQUEST_PATH'] ||= [env['SCRIPT_NAME'], env['PATH_INFO']].join

            status, headers, body = @app.call(env)
            begin
              res.status = status.to_i
              headers.each { |k, vs|
                next if k.downcase == 'rack.hijack'

                if k.downcase == 'set-cookie'
                  res.cookies.concat vs.split("\n")
                else
                  # Since WEBrick won't accept repeated headers,
                  # merge the values per RFC 1945 section 4.2.
                  res[k] = vs.split("\n").join(", ")
                end
              }

              io_lambda = headers['rack.hijack']
              if io_lambda
                rd, wr = IO.pipe
                res.body = rd
                res.chunked = true
                io_lambda.call wr
              elsif body.respond_to?(:to_path)
                res.body = ::File.open(body.to_path, 'rb')
              else
                r, w = IO.pipe
                w.write('')
                Thread.new {
                  body.each { |part|
                    w.write(part)
                  }
                  w.close
                }
                res.body = r
              end
            ensure
              body.close  if body.respond_to? :close
            end
          end
        end

      end

    end
  end
end
